import unittest
from itertools import combinations

from phevaluator.tables import FLUSH


class TestFlushTable(unittest.TestCase):
    TABLE = [0] * len(FLUSH)
    VISIT = [0] * len(FLUSH)
    CUR_RANK = 1

    CACHE = []
    BINARIES = []

    @classmethod
    def setUpClass(cls):
        cls.mark_straight()
        cls.mark_four_of_a_kind()
        cls.mark_full_house()
        cls.mark_non_straight()

    @classmethod
    def gen_binary(cls, highest, k, n):
        if k == 0:
            cls.BINARIES.append(cls.CACHE[:])
        else:
            for i in range(highest, -1, -1):
                cls.CACHE.append(i)
                cls.gen_binary(i - 1, k - 1, n)
                cls.CACHE.remove(i)

    @classmethod
    def mark_straight(cls):
        for highest in range(12, 3, -1):  # From Ace to 6
            # k=5 case for base
            base = [highest - i for i in range(5)]
            base_idx = 0
            for pos in base:
                base_idx += 1 << pos
            cls.TABLE[base_idx] = cls.CUR_RANK
            cls.VISIT[base_idx] = 1
            cls.mark_six_to_nine(base, base_idx)

            # setting up for the next loop
            cls.CUR_RANK += 1

        # Five High Straight Flush
        base = [12, 3, 2, 1, 0]
        base_idx = 0
        for pos in base:
            base_idx += 1 << pos
        cls.TABLE[base_idx] = cls.CUR_RANK
        cls.VISIT[base_idx] = 1
        cls.mark_six_to_nine(base, base_idx)
        cls.CUR_RANK += 1

    @classmethod
    def mark_non_straight(cls):
        cls.gen_binary(12, 5, 5)
        for base in cls.BINARIES:
            base_idx = 0
            for pos in base:
                base_idx += 1 << pos

            if cls.VISIT[base_idx] > 0:
                continue

            cls.TABLE[base_idx] = cls.CUR_RANK
            cls.VISIT[base_idx] = 1
            cls.mark_six_to_nine(base, base_idx)

            # setting up for the next loop
            cls.CUR_RANK += 1

    @classmethod
    def mark_six_to_nine(cls, base, base_idx):
        # k=6-9 cases
        pos_candidates = [i for i in range(13) if i not in base]
        for r in [1, 2, 3, 4]:  # Need to select additional cards
            for cb in combinations(pos_candidates, r):
                idx = base_idx
                for pos in cb:
                    idx += 1 << pos

                if cls.VISIT[idx] > 0:
                    continue

                cls.TABLE[idx] = cls.CUR_RANK
                cls.VISIT[idx] = 1

    @classmethod
    def mark_four_of_a_kind(cls):
        # Four of a kind
        # The rank of the four cards: 13C1
        # The rank of the other card: 12C1
        cls.CUR_RANK += 13 * 12

    @classmethod
    def mark_full_house(cls):
        # Full house
        # The rank of the cards of three of a kind: 13C1
        # The rank of the cards of a pair: 12C1
        cls.CUR_RANK += 13 * 12

    def test_flush_table(self):
        self.assertListEqual(self.TABLE, FLUSH)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

from phevaluator import Card


class TestCard(unittest.TestCase):
    # lowercase name, capitalcase name, value
    testcases = [
        ["2c", "2C", 0],
        ["2d", "2D", 1],
        ["2h", "2H", 2],
        ["2s", "2S", 3],
        ["Tc", "TC", 32],
        ["Ac", "AC", 48],
    ]

    def test_card_equality(self):
        for name, capital_name, number in self.testcases:
            # equality between cards
            # e.g. Card("2c") == Card(0)
            self.assertEqual(Card(name), Card(number))
            # e.g. Card("2C") == Card(0)
            self.assertEqual(Card(capital_name), Card(number))
            # e.g. Card(Card(0)) == Card(0)
            self.assertEqual(Card(Card(number)), Card(number))

            # equality between Card and int
            self.assertEqual(Card(number), number)  # e.g. Card(0) == 0

    def test_card_immutability(self):
        # Once a Card is assigned or constructed from another Card, it's not affected by any changes to source variable
        c_source = Card(1)
        c_assign = c_source
        c_construct = Card(c_source)

        c_source += 1

        self.assertNotEqual(c_source, Card(1))
        self.assertEqual(c_assign, Card(1))
        self.assertEqual(c_construct, Card(1))

    def test_card_describe(self):
        for name, capital_name, number in self.testcases:
            rank, suit = name
            c_name = Card(name)
            c_capital_name = Card(capital_name)
            c_number = Card(number)
            c_construct = Card(c_number)

            # Card("2c").describe_rank() == "2"
            self.assertEqual(c_name.describe_rank(), rank)
            # Card("2c").describe_suit() == "c"
            self.assertEqual(c_name.describe_suit(), suit)

            # Card("2c").describe_card() == "2c"
            self.assertEqual(c_name.describe_card(), name)

            # Card("2C").describe_card() == "2c"
            self.assertEqual(c_capital_name.describe_card(), name)
            # Card("2C").describe_card() != "2C"
            self.assertNotEqual(c_capital_name.describe_card(), capital_name)

            # Card(0).describe_card() == "2c"
            self.assertEqual(c_number.describe_card(), name)

            # Card(Card(0)).describe_card() == "2c"
            self.assertEqual(c_construct.describe_card(), name)

            # str(Card("2c")) == "2c"
            self.assertEqual(str(c_name), name)
            # repr(Card("2c")) == 'Card("2c")'
            self.assertEqual(repr(c_name), f'Card("{name}")')



import json
import os
import unittest

from phevaluator import Card, evaluate_cards, evaluate_omaha_cards

CARDS_FILE_5 = os.path.join(os.path.dirname(__file__), "cardfiles/5cards.json")
CARDS_FILE_6 = os.path.join(os.path.dirname(__file__), "cardfiles/6cards.json")
CARDS_FILE_7 = os.path.join(os.path.dirname(__file__), "cardfiles/7cards.json")


class TestEvaluator(unittest.TestCase):
    def test_example(self):
        p1 = evaluate_cards("9c", "4c", "4s", "9d", "4h", "Qc", "6c")
        p2 = evaluate_cards("9c", "4c", "4s", "9d", "4h", "2c", "9h")

        self.assertEqual(p1, 292)
        self.assertEqual(p2, 236)
        self.assertLess(p2, p1)

    def test_omaha_example(self):
        # fmt: off
        p1 = evaluate_omaha_cards(
            "4c", "5c", "6c", "7s", "8s", # community cards
            "2c", "9c", "As", "Kd",       # player hole cards
        )

        p2 = evaluate_omaha_cards(
            "4c", "5c", "6c", "7s", "8s", # community cards
            "6s", "9s", "Ts", "Js",       # player hole cards
        )
        # fmt: on

        self.assertEqual(p1, 1578)
        self.assertEqual(p2, 1604)

    def test_5cards(self):
        with open(CARDS_FILE_5, "r") as read_file:
            hand_dict = json.load(read_file)
            for key, value in hand_dict.items():
                self.assertEqual(evaluate_cards(*key.split()), value)

    def test_6cards(self):
        with open(CARDS_FILE_6, "r") as read_file:
            hand_dict = json.load(read_file)
            for key, value in hand_dict.items():
                self.assertEqual(evaluate_cards(*key.split()), value)

    def test_7cards(self):
        with open(CARDS_FILE_7, "r") as read_file:
            hand_dict = json.load(read_file)
            for key, value in hand_dict.items():
                self.assertEqual(evaluate_cards(*key.split()), value)

    def test_evaluator_interface(self):
        # int, str and Card can be passed to evaluate_cards()
        p1 = evaluate_cards(1, 2, 3, 32, 48)
        p2 = evaluate_cards("2d", "2h", "2s", "Tc", "Ac")
        p3 = evaluate_cards("2D", "2H", "2S", "TC", "AC")
        p4 = evaluate_cards(Card("2d"), Card("2h"), Card("2s"), Card("Tc"), Card("Ac"))
        p5 = evaluate_cards(1, "2h", "2S", Card(32), Card("Ac"))

        self.assertEqual(p1, p2)
        self.assertEqual(p1, p3)
        self.assertEqual(p1, p4)
        self.assertEqual(p1, p5)



import unittest
from itertools import combinations, combinations_with_replacement, permutations

from phevaluator.hash import hash_quinary
from phevaluator.tables import NO_FLUSH_5


class BaseTestNoFlushTable(unittest.TestCase):
    TABLE = NotImplementedError
    VISIT = NotImplementedError
    NUM_CARDS = NotImplementedError

    @classmethod
    def setUpClass(cls):
        cls.CACHE = []
        cls.USED = [0] * 13
        cls.QUINARIES = []

        cls.CACHE_ADDITIONAL = []
        cls.USED_ADDITIONAL = [0] * 13
        cls.QUINARIES_ADDITIONAL = []

        # Straight Flushes are not in this table
        cls.mark_four_of_a_kind()
        cls.mark_full_house()
        # Flushes are not in this table
        cls.mark_straight()
        cls.mark_three_of_a_kind()
        cls.mark_two_pair()
        cls.mark_one_pair()
        cls.mark_high_card()

    @staticmethod
    def quinary_permutations(n):
        return permutations(range(13)[::-1], n)

    @staticmethod
    def quinary_combinations(n):
        return combinations(range(13)[::-1], n)

    @staticmethod
    def quinary_combinations_with_replacement(n):
        return combinations_with_replacement(range(13)[::-1], n)

    @classmethod
    def gen_quinary(cls, ks, cur, additional):
        if cur == len(ks):
            cls.get_additional(additional)
            cls.QUINARIES.append((cls.CACHE[:], cls.QUINARIES_ADDITIONAL[:]))
            cls.QUINARIES_ADDITIONAL = []
        else:
            for i in range(12, -1, -1):
                if cls.USED[i] > 0:
                    continue
                cls.CACHE.append(i)
                cls.USED[i] = ks[cur]
                cls.gen_quinary(ks, cur + 1, additional)
                cls.CACHE.pop(-1)
                cls.USED[i] = 0

    @classmethod
    def get_additional(cls, n):
        if n == 0:
            cls.QUINARIES_ADDITIONAL.append(cls.CACHE_ADDITIONAL[:])
        else:
            for i in range(12, -1, -1):
                if cls.USED[i] + cls.USED_ADDITIONAL[i] >= 4:
                    continue
                cls.CACHE_ADDITIONAL.append(i)
                cls.USED_ADDITIONAL[i] += 1
                cls.get_additional(n - 1)
                cls.CACHE_ADDITIONAL.pop(-1)
                cls.USED_ADDITIONAL[i] -= 1

    @classmethod
    def mark_template(cls, ks):
        cls.gen_quinary(ks, 0, cls.NUM_CARDS - 5)
        for base, additionals in cls.QUINARIES:
            hand = [0] * 13
            for i, k in enumerate(ks):
                hand[base[i]] = k
            base_rank = NO_FLUSH_5[hash_quinary(hand, 5)]
            for additional in additionals:
                for i in additional:

                    hand[i] += 1

                hash_ = hash_quinary(hand, cls.NUM_CARDS)

                for i in additional:
                    hand[i] -= 1

                if cls.VISIT[hash_] > 0:
                    continue
                cls.TABLE[hash_] = base_rank
                cls.VISIT[hash_] = 1

        cls.QUINARIES = []

    @classmethod
    def mark_four_of_a_kind(cls):
        cls.mark_template((4, 1))

    @classmethod
    def mark_full_house(cls):
        cls.mark_template((3, 2))

    @classmethod
    def mark_three_of_a_kind(cls):
        cls.mark_template((3, 1, 1))

    @classmethod
    def mark_two_pair(cls):
        cls.mark_template((2, 2, 1))

    @classmethod
    def mark_one_pair(cls):
        for paired_card in range(13)[::-1]:
            for other_cards in cls.quinary_combinations(cls.NUM_CARDS - 2):
                if paired_card in other_cards:
                    continue
                hand = [0] * 13
                hand[paired_card] = 2
                for i in range(3):
                    hand[other_cards[i]] = 1
                base_hash = hash_quinary(hand, 5)
                base_rank = NO_FLUSH_5[base_hash]
                for i in range(3, cls.NUM_CARDS - 2):
                    hand[other_cards[i]] = 1
                hash_ = hash_quinary(hand, cls.NUM_CARDS)
                if cls.VISIT[hash_] == 0:
                    cls.VISIT[hash_] = 1
                    cls.TABLE[hash_] = base_rank

    @classmethod
    def mark_high_card(cls):
        for base in cls.quinary_combinations(cls.NUM_CARDS):
            hand = [0] * 13
            for i in range(5):
                hand[base[i]] = 1
            base_hash = hash_quinary(hand, 5)
            base_rank = NO_FLUSH_5[base_hash]
            for i in range(5, cls.NUM_CARDS):
                hand[base[i]] = 1
            hash_ = hash_quinary(hand, cls.NUM_CARDS)
            if cls.VISIT[hash_] == 0:
                cls.VISIT[hash_] = 1
                cls.TABLE[hash_] = base_rank

    @classmethod
    def mark_straight(cls):
        hands = []
        for lowest in range(9)[::-1]:  # From 10 to 2
            hand = [0] * 13
            for i in range(lowest, lowest + 5):
                hand[i] = 1
            hands.append(hand)
        # Five High Straight Flush
        base = [12, 3, 2, 1, 0]
        hand = [0] * 13
        for i in base:
            hand[i] = 1
        hands.append(hand)

        for hand in hands:
            base_rank = NO_FLUSH_5[hash_quinary(hand, 5)]
            for additional in cls.quinary_combinations_with_replacement(
                cls.NUM_CARDS - 5
            ):
                for i in additional:
                    hand[i] += 1

                hash_ = hash_quinary(hand, cls.NUM_CARDS)
                if cls.VISIT[hash_] == 0:
                    cls.TABLE[hash_] = base_rank
                    cls.VISIT[hash_] = 1

                for i in additional:
                    hand[i] -= 1



import unittest
from itertools import combinations, permutations

from phevaluator.hash import hash_quinary
from phevaluator.tables import NO_FLUSH_5


class TestNoFlush5Table(unittest.TestCase):
    TABLE = [0] * len(NO_FLUSH_5)
    VISIT = [0] * len(NO_FLUSH_5)
    CUR_RANK = 1
    NUM_CARDS = 5

    @classmethod
    def setUpClass(cls):
        cls.mark_straight_flush()
        cls.mark_four_of_a_kind()
        cls.mark_full_house()
        cls.mark_flush()
        cls.mark_straight()
        cls.mark_three_of_a_kind()
        cls.mark_two_pair()
        cls.mark_one_pair()
        cls.mark_high_card()

    @staticmethod
    def quinaries(n):
        return permutations(range(13)[::-1], n)

    @staticmethod
    def quinaries_without_duplication():
        return combinations(range(13)[::-1], 5)

    @classmethod
    def mark_four_of_a_kind(cls):
        # Order 13C2 lexicographically
        for base in cls.quinaries(2):
            hand = [0] * 13
            hand[base[0]] = 4
            hand[base[1]] = 1
            hash_ = hash_quinary(hand, cls.NUM_CARDS)
            cls.TABLE[hash_] = cls.CUR_RANK
            cls.VISIT[hash_] = 1
            cls.CUR_RANK += 1

    @classmethod
    def mark_full_house(cls):
        for base in cls.quinaries(2):
            hand = [0] * 13
            hand[base[0]] = 3
            hand[base[1]] = 2
            hash_ = hash_quinary(hand, cls.NUM_CARDS)
            cls.TABLE[hash_] = cls.CUR_RANK
            cls.VISIT[hash_] = 1
            cls.CUR_RANK += 1

    @classmethod
    def mark_straight(cls):
        for lowest in range(9)[::-1]:  # From 10 to 2
            hand = [0] * 13
            for i in range(lowest, lowest + 5):
                hand[i] = 1
            hash_ = hash_quinary(hand, cls.NUM_CARDS)
            cls.TABLE[hash_] = cls.CUR_RANK
            cls.VISIT[hash_] = 1
            cls.CUR_RANK += 1

        # Five High Straight Flush
        base = [12, 3, 2, 1, 0]
        hand = [0] * 13
        for pos in base:
            hand[pos] = 1

        hash_ = hash_quinary(hand, cls.NUM_CARDS)
        cls.TABLE[hash_] = cls.CUR_RANK
        cls.VISIT[hash_] = 1
        cls.CUR_RANK += 1

    @classmethod
    def mark_three_of_a_kind(cls):
        for base in cls.quinaries(3):
            hand = [0] * 13
            hand[base[0]] = 3
            hand[base[1]] = 1
            hand[base[2]] = 1
            hash_ = hash_quinary(hand, cls.NUM_CARDS)
            if cls.VISIT[hash_] == 0:
                cls.TABLE[hash_] = cls.CUR_RANK
                cls.VISIT[hash_] = 1
                cls.CUR_RANK += 1

    @classmethod
    def mark_two_pair(cls):
        for base in cls.quinaries(3):
            hand = [0] * 13
            hand[base[0]] = 2
            hand[base[1]] = 2
            hand[base[2]] = 1
            hash_ = hash_quinary(hand, cls.NUM_CARDS)
            if cls.VISIT[hash_] == 0:
                cls.TABLE[hash_] = cls.CUR_RANK
                cls.VISIT[hash_] = 1
                cls.CUR_RANK += 1

    @classmethod
    def mark_one_pair(cls):
        for base in cls.quinaries(4):
            hand = [0] * 13
            hand[base[0]] = 2
            hand[base[1]] = 1
            hand[base[2]] = 1
            hand[base[3]] = 1
            hash_ = hash_quinary(hand, cls.NUM_CARDS)
            if cls.VISIT[hash_] == 0:
                cls.TABLE[hash_] = cls.CUR_RANK
                cls.VISIT[hash_] = 1
                cls.CUR_RANK += 1

    @classmethod
    def mark_high_card(cls):
        for base in cls.quinaries_without_duplication():
            hand = [0] * 13
            hand[base[0]] = 1
            hand[base[1]] = 1
            hand[base[2]] = 1
            hand[base[3]] = 1
            hand[base[4]] = 1
            hash_ = hash_quinary(hand, cls.NUM_CARDS)
            if cls.VISIT[hash_] == 0:
                cls.TABLE[hash_] = cls.CUR_RANK
                cls.VISIT[hash_] = 1
                cls.CUR_RANK += 1

    @classmethod
    def mark_straight_flush(cls):
        # A-5 High Straight Flush: 10
        cls.CUR_RANK += 10

    @classmethod
    def mark_flush(cls):
        # Selecting 5 cards in 13: 13C5
        # Need to exclude straight: -10
        cls.CUR_RANK += int(13 * 12 * 11 * 10 * 9 / (5 * 4 * 3 * 2)) - 10

    def test_noflush5_table(self):
        self.assertListEqual(self.TABLE, NO_FLUSH_5)


if __name__ == "__main__":
    unittest.main()



import unittest
from itertools import combinations

from phevaluator import (
    Card,
    _evaluate_cards,
    _evaluate_omaha_cards,
    evaluate_omaha_cards,
    sample_cards,
)


def evaluate_omaha_exhaustive(community_cards: list[int], hole_cards: list[int]) -> int:
    """Evaluate omaha cards with `_evaluate_cards`."""
    best_rank = min(
        _evaluate_cards(c1, c2, c3, h1, h2)
        for c1, c2, c3 in combinations(community_cards, 3)
        for h1, h2 in combinations(hole_cards, 2)
    )
    return best_rank


class TestEvaluatorOmaha(unittest.TestCase):
    def test_omaha(self):
        """Compare the evaluation between `_evaluate_omaha_cards` and `_evaluate_cards`"""
        total = 10000
        for _ in range(total):
            cards = sample_cards(9)
            community_cards = cards[:5]
            hole_cards = cards[5:]
            with self.subTest(cards):
                self.assertEqual(
                    _evaluate_omaha_cards(community_cards, hole_cards),
                    evaluate_omaha_exhaustive(community_cards, hole_cards),
                )

    def test_evaluator_interface(self):
        # int, str and Card can be passed to evaluate_omaha_cards()
        # fmt: off
        rank1 = evaluate_omaha_cards(
            48, 49, 47, 43, 35, # community cards
            51, 50, 39, 34      # hole cards
        )
        rank2 = evaluate_omaha_cards(
            "Ac", "Ad", "Ks", "Qs", "Ts", # community cards
            "As", "Ah", "Js", "Th"        # hole cards
        )
        rank3 = evaluate_omaha_cards(
            "AC", "AD", "KS", "QS", "TS", # community cards
            "AS", "AH", "JS", "TH"        # hole cards
        )
        rank4 = evaluate_omaha_cards(
            Card("Ac"), Card("Ad"), Card("Ks"), Card("Qs"), Card("Ts"), # community cards
            Card("As"), Card("Ah"), Card("Js"), Card("Th")              # hole cards
        )
        rank5 = evaluate_omaha_cards(
            48, "Ad", "KS", Card(43), Card("Ts"), # community cards
            Card("AS"), 50, "Js", "TH"            # hole cards
        )
        # fmt: on
        self.assertEqual(rank1, rank2)
        self.assertEqual(rank1, rank3)
        self.assertEqual(rank1, rank4)
        self.assertEqual(rank1, rank5)



from typing import Iterable, Union

from .card import Card
from .hash import hash_quinary
from .tables import (
    BINARIES_BY_ID,
    FLUSH,
    NO_FLUSH_5,
    NO_FLUSH_6,
    NO_FLUSH_7,
    SUITBIT_BY_ID,
    SUITS,
)

MIN_CARDS = 5
MAX_CARDS = 7

NO_FLUSHES = {5: NO_FLUSH_5, 6: NO_FLUSH_6, 7: NO_FLUSH_7}


def evaluate_cards(*cards: Iterable[Union[int, str, Card]]) -> int:
    cards = list(map(Card, cards))
    hand_size = len(cards)

    if (not (MIN_CARDS <= hand_size <= MAX_CARDS)) or (hand_size not in NO_FLUSHES):
        raise ValueError(
            f"The number of cards must be between {MIN_CARDS} and {MAX_CARDS}."
            f"passed size: {hand_size}"
        )

    NO_FLUSH = NO_FLUSHES[hand_size]

    suit_hash = 0
    for card in cards:
        suit_hash += SUITBIT_BY_ID[card]

    flush_suit = SUITS[suit_hash] - 1

    if flush_suit != -1:
        hand_binary = 0

        for card in cards:
            if card % 4 == flush_suit:
                hand_binary |= BINARIES_BY_ID[card]

        return FLUSH[hand_binary]

    hand_quinary = [0] * 13
    for card in cards:
        hand_quinary[card // 4] += 1

    return NO_FLUSH[hash_quinary(hand_quinary, hand_size)]



import unittest

from phevaluator.tables import NO_FLUSH_7

from .utils import BaseTestNoFlushTable


class TestNoFlush7Table(BaseTestNoFlushTable):
    TOCOMPARE = NO_FLUSH_7
    TABLE = [0] * len(TOCOMPARE)
    VISIT = [0] * len(TOCOMPARE)
    NUM_CARDS = 7

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def test_noflush7_table(self):
        self.assertListEqual(self.TABLE, self.TOCOMPARE)


if __name__ == "__main__":
    unittest.main()



import unittest
from itertools import combinations_with_replacement

from phevaluator.tables import CHOOSE, DP, SUITS


class TestSuitsTable(unittest.TestCase):
    DP = [0] * len(SUITS)

    @classmethod
    def setUpClass(cls):
        for k in [5, 6, 7, 8, 9]:
            cls.update_k(cls.DP, k)

    @staticmethod
    def update_k(table, k):
        iterable = list(range(0, k + 1))
        cbs = combinations_with_replacement(iterable, 3)

        for cb in cbs:
            # cb is in lexicographically sorted order
            cnts = (cb[0], cb[1] - cb[0], cb[2] - cb[1], k - cb[2])
            for suit, cnt in enumerate(cnts):
                if cnt >= 5:
                    idx = (
                        0x1 * cnts[0] + 0x8 * cnts[1] + 0x40 * cnts[2] + 0x200 * cnts[3]
                    )

                    # TODO: Need to check these cases:
                    #  There exist three cases that idxes are same.
                    #  For two different cnts in case of k=9. The
                    #  cases are 72, 520, 576.
                    if idx in [72, 520, 576] and SUITS[idx] != suit + 1:
                        continue

                    table[idx] = suit + 1

    def test_suits_table(self):
        self.assertListEqual(self.DP, SUITS)


class TestChooseTable(unittest.TestCase):
    DP = [[0] * len(CHOOSE[idx]) for idx in range(len(CHOOSE))]
    VISIT = [[0] * len(CHOOSE[idx]) for idx in range(len(CHOOSE))]

    @classmethod
    def setUpClass(cls):
        for row in range(len(CHOOSE)):
            for col in range(len(CHOOSE[row])):
                cls.nCr(row, col)

    @classmethod
    def nCr(cls, n, r):
        if n < r:
            return 0
        elif r == 0:
            cls.DP[n][r] = 1
            return 1
        else:
            if cls.VISIT[n][r] == 0:
                cls.DP[n][r] = cls.nCr(n - 1, r) + cls.nCr(n - 1, r - 1)
                cls.VISIT[n][r] = 1
            return cls.DP[n][r]

    def test_choose_table(self):
        self.assertListEqual(self.DP, CHOOSE)


class TestDpTable(unittest.TestCase):
    DP = [[[0] * len(DP[i][j]) for j in range(len(DP[i]))] for i in range(len(DP))]

    @classmethod
    def setUpClass(cls):
        cls.fill_table()

    @classmethod
    def fill_table(cls):
        # Recursion formula:
        # dp[l][i][j] = dp[l-1][i][j] + dp[1][i][j-l+1]
        #
        # We need base cases of dp[1][i][j] to calculate.
        #
        # Base cases:
        # dp[1][i][j] is something like combination with
        # replacement(iHj), but each bag cannot be bigger than 4.
        # (1) dp[1][1][j] = 1 for 0 <= j <= 4, 0 for j > 4
        #     dp[1][0][j] = 0 for i = 0 (invalid)
        # (2) dp[1][i>1][j] = SUM { dp[1][i-1][j-q] }
        #     for q from 0 to 4 where j-q >= 0.
        #     This is like setting the most left number to q.
        # We need (2) because of the restriction.

        # Make base cases
        for j in range(0, 5):
            cls.DP[1][1][j] = 1
        for i in range(2, 14):
            for j in range(10):
                for q in range(5):
                    if j - q >= 0:
                        cls.DP[1][i][j] += cls.DP[1][i - 1][j - q]

        # Make recursion
        for l in range(2, 5):
            for i in range(14):
                for j in range(10):
                    cls.DP[l][i][j] = cls.DP[l - 1][i][j]
                    if j - l + 1 >= 0:
                        cls.DP[l][i][j] += cls.DP[1][i][j - l + 1]

    def test_dp_table(self):
        self.assertListEqual(self.DP, DP)


if __name__ == "__main__":
    unittest.main()



from typing import Iterable

from .tables import CHOOSE, DP


def hash_quinary(q: Iterable[int], k: int) -> int:
    sum_numb = 0
    length = len(q)

    for rank, cnt in enumerate(q):
        if cnt == 0:
            continue

        sum_numb += DP[cnt][length - rank - 1][k]

        k -= cnt
        if k <= 0:
            break

    return sum_numb


def hash_binary(binary: int, k: int) -> int:
    sum_numb = 0
    length = 15

    for rank in range(length):
        if binary & (1 << rank):
            if length - rank - 1 >= k:
                sum_numb += CHOOSE[length - rank - 1][k]

            k -= 1
            if k == 0:
                break

    return sum_numb
